var path = require('path');

require.paths.unshift(path.dirname(__dirname));
require.paths.unshift(__dirname);

var Config = require('config/NodeRunnerConfiguration');
var sys = require('sys');
var exec = require('child_process').exec;
var http = require('http');
var https = require('https');
var fs = require('fs');
var url = require('url');

var Runner = function() {
	var child = null;
	var portsInUse = [];
	var current = 0;
	var port = null;
	var httpHandler = null;
	var httpsHandler = null;
	var httpsOptions = null;
	var onCreateHandler = null;
	var serverOptions = null;
	var errorMsg = "The following ports are in use ";

	/** 
	 * Spawns a new server 
	 */
	this.run = function(server) {
		if (typeof server.httpHandler !== 'function' && server.httpsHandler !== 'function') {
			throw new TypeError("Please specify a valid callback handler to be invoked after creating the server...");
		}

		console.log(server.serverOptions);

		if (typeof server.serverOptions === 'undefined') {
			throw new TypeError('Please specify the server config: port:https, etc...');
		} else if (!server.serverOptions.ports || typeof server.serverOptions.ports !== 'object') {
			throw new TypeError('Please specify one or more ports to connect to');
		}

		httpHandler = server.httpHandler;
		serverOptions = server.serverOptions;

		if (server.httpsHandler) {
			if (typeof server.serverOptions.cert === 'undefined') {
				throw new TypeError('Please specify the HTTPS config options: certificate and key');
			} else if (typeof server.serverOptions.cert.key === 'undefined') {
				throw new TypeError('Please specify the HTTPS config option: Certificate key file');
			} else if (typeof server.serverOptions.cert.cert === 'undefined') {
				throw new TypeError('Please specify the HTTPS config option: Certificate file');
			}

			httpsHandler = server.httpsHandler;
		}

		onCreateHandler = server.onCreateHandler;

		test();
	};

	/** 
	 * Test the specified ports are within the
	 * allowed range.
	 */
	var test = function() {
		sys.puts("------------------------------------------------------------------------");
		sys.puts("Preprocessing... Checking port range specified...");
		sys.puts("------------------------------------------------------------------------");
		serverOptions.ports.forEach(function(val, index, array) {
			sys.puts('......... Checking if port ' + val.port + ' is within the allowed range');
			if (val.port < Config.allowedPortRange.min || val.port > Config.allowedPortRange.max) {
				error("Invalid port specified. Port # should be between " + Config.allowedPortRange.min + " and " + Config.allowedPortRange.max);
				return;
			} else {
				success("The port " + val.port + ' is within the allowed range...');
			}
		});

		sys.puts("------------------------------------------------------------------------");

		_process(null);
	};

	/**
	 * If using HTTP, check to see if 
	 * something is already running on that port...
	 */
	var testHTTP = function(port, cb) {
		var options = {
			'host': port.hostname,
			'port': port.port
		};

		var self = this;
		http.get(options, function(res) {
			portsInUse.push(port.port);
			cb.apply(self, [true]);
		}).on('error', function(e) {
			cb.apply(self, [false]);
		});
	};

	var testHTTPS = function(port, cb) {
		var self = this;
		var options = {
			host: port.hostname,
			port: port.port,
			path: '/',
			method: 'GET'
		};

		var req = https.request(options, function(res) {
			sys.puts("statusCode: ", res.statusCode);
			sys.puts("headers: ", res.headers);
			portsInUse.push(port.port);
			cb.apply(self, [true]);
		});

		req.on('error', function(e) {
			cb.apply(self, [false]);
		});
	};

	var testPorts = function(port, _cb) {
		testHTTPS(port, function(_httpsError) {
			testHTTP(port, function(_httpError) {
				current++;
				if (_httpsError === true && _httpError === true) {
					_cb(true);
				} else {
					info(port.hostname + ":" + port.port + " is available for use...");
					spawn(port);
					_cb(false);
				}
			});
		});
	};

	var getPort = function() {
		return serverOptions.ports[current];
	};

	var getPortSize = function() {
		return serverOptions.ports.length;
	};

	var _process = function(_error) {
		if (current >= getPortSize()) {
			if (portsInUse.lenght > 0) {
				for (var i in portsInUse) {
					errorMsg += portsInUse[i] + ' ';
				}
				error(errorMsg);
			}

			return;
		}

		var port = serverOptions.ports[current];
		port.hostname = port.hostname || 'localhost';

		info("Processing next port... " + port.port);

		testPorts(port, _process);
	};

	var success = function(msg) {
		sys.puts('[SUCCESS] ' + msg);
	};
	
	var info = function(msg) {
		sys.puts('[INFO] ' + msg);
	};

	var error = function(message) {
		sys.puts("------------------------------------------------------------------------");
		sys.puts("** [ERROR] There were one or more errors launching the server instances.");
		sys.puts("** [ERROR] Please review the reported errors and try again.");
		sys.puts("------------------------------------------------------------------------");
		sys.puts(message);
		sys.puts("** [ERROR] Exiting...");

		process.exit(1);
	};

	var spawn = function(port) {
		if (typeof port === 'undefined') {
			throw new TypeError('Invalid port specified');
		}

		port.hostname = port.hostname || 'localhost';
		var server = null;
		if (port.https === 'true' || port.https === true) {
			if (typeof httpsHandler !== 'function') {
				throw new TypeError('Invalid HTTPS handler specified');
			}

			server = https.createServer({
				key: fs.readFileSync(serverOptions.cert.key),
				cert: fs.readFileSync(serverOptions.cert.cert)
			},
			httpsHandler);

			server.listen(port);

			success("Spawned new HTTPS Server on https://" + port.hostname + ":" + port.port);

		} else if (port.https === 'false' || port.https === false) {
			if (typeof httpHandler !== 'function') {
				throw new TypeError('Invalid HTTP handler specified');
			}

			server = http.createServer(httpHandler);
			server.listen(port.port, port.hostname);
			success("Spawned new HTTP Server on http://" + port.hostname + ":" + port.port);
		}

		if (typeof onCreateHandler !== 'undefined') {
			onCreateHandler(server, {
				'port': port.port,
				'hostname': port.hostname
			});
		}
	};

	return this;
};

module.exports.Runner = new Runner();
